


#include "ClothActor.h"

#include "ClothSimulationComponent.h"
#include "Providers/RuntimeMeshProviderStatic.h"

// Sets default values
AClothActor::AClothActor()
{
 	// Set this actor to call Tick() every frame.  You can turn this off to improve performance if you don't need it.
	PrimaryActorTick.bCanEverTick = true;
	SetMobility(EComponentMobility::Movable);
	SetRuntimeMeshMobility(ERuntimeMeshMobility::Movable);
	
	StaticProvider = CreateDefaultSubobject<URuntimeMeshProviderStatic>(TEXT("RuntimeMeshProvicerStatic"));
	StaticProvider->SetShouldSerializeMeshData(false);

	ClothSimulation = CreateDefaultSubobject<UClothSimulationComponent>(TEXT("ClothSimulation"));
	ClothSimulation->SetupAttachment(RootComponent);
}

void AClothActor::GenerateGrid()
{
	MeshData.Reset();

	int32 VertexIndex = 0;
	for (int32 Y = 0; Y <= GridSize; ++Y)
	{
		for (int32 X = 0; X <= GridSize; ++X, ++VertexIndex)
		{
			// Each vertex gets a position from the simulation and a texture coordinate.
			FVector3f Position = ClothSimulation->GetNodePosition(VertexIndex);
			MeshData.Positions.Add(Position);

			const float U = static_cast<float>(X) / GridSize;
			const float V = static_cast<float>(Y) / GridSize;
			MeshData.TexCoords.Add({U, V});

			// Add an empty tangent and normal to be filled later.
			MeshData.Tangents.Add(FVector3f::ZeroVector, FVector3f::ZeroVector);

			// And add white.
			MeshData.Colors.Add(FColor::White);

			// When we have enough vertices and are not at the start of a line, we can start
			// generating triangles from them. Each vertex adds 2 triangles to the grid (1 quad).
			if (X > 0 && Y > 0)
			{
				const int32 TopRightIndex = VertexIndex - GridSize - 1;
				const int32 TopLeftIndex = VertexIndex - GridSize - 2;
				const int32 BottomLeftIndex = VertexIndex - 1;
				const int32 BottomRightIndex = VertexIndex;
				
				// Now create two triangles from these four vertices.
				// The order of these (clockwise/counter-clockwise) dictates which way the triangle will face.
				MeshData.Triangles.AddTriangle(BottomLeftIndex, TopRightIndex, TopLeftIndex);
				MeshData.Triangles.AddTriangle(BottomLeftIndex, BottomRightIndex, TopRightIndex);

				const FVector3f NormalCurrent = FVector3f::CrossProduct(
					MeshData.Positions.GetPosition(BottomLeftIndex) - MeshData.Positions.GetPosition(TopLeftIndex),
					MeshData.Positions.GetPosition(TopLeftIndex) - MeshData.Positions.GetPosition(TopRightIndex)).GetSafeNormal();

				MeshData.Tangents.SetNormal(TopLeftIndex, NormalCurrent);
				MeshData.Tangents.SetNormal(TopRightIndex, NormalCurrent);
				MeshData.Tangents.SetNormal(BottomLeftIndex, NormalCurrent);
				MeshData.Tangents.SetNormal(BottomRightIndex, NormalCurrent);
			}
		}
	}
}

void AClothActor::GenerateMesh()
{
	GetRuntimeMeshComponent()->Initialize(StaticProvider);
	StaticProvider->ClearSection(0, 0);
	StaticProvider->SetupMaterialSlot(0, TEXT("CylinderMaterial"), Material);

	MeshProperties.UpdateFrequency = ERuntimeMeshUpdateFrequency::Frequent;
	MeshProperties.MaterialSlot = 0;

	GenerateGrid();
	StaticProvider->CreateSection(0, 0, MeshProperties, MeshData, false);
}

void AClothActor::UpdateMesh()
{
	int32 VertexIndex = 0;
	for (int32 Y = 0; Y <= GridSize; ++Y)
	{
		for (int32 X = 0; X <= GridSize; ++X, ++VertexIndex)
		{
			// Each vertex gets its position updated from the simulation.
			FVector3f Position = ClothSimulation->GetNodePosition(VertexIndex);
			MeshData.Positions.SetPosition(VertexIndex, Position);

			// We also need to update normals for each of the triangles in the grid.
			if (X > 0 && Y > 0)
			{
				const int32 TopRightIndex = VertexIndex - GridSize - 1;
				const int32 TopLeftIndex = VertexIndex - GridSize - 2;
				const int32 BottomLeftIndex = VertexIndex - 1;
				const int32 BottomRightIndex = VertexIndex;

				const FVector3f NormalCurrent = FVector3f::CrossProduct(
					MeshData.Positions.GetPosition(BottomLeftIndex) - MeshData.Positions.GetPosition(TopLeftIndex),
					MeshData.Positions.GetPosition(TopLeftIndex) - MeshData.Positions.GetPosition(TopRightIndex)).GetSafeNormal();

				MeshData.Tangents.SetNormal(TopLeftIndex, NormalCurrent);
				MeshData.Tangents.SetNormal(TopRightIndex, NormalCurrent);
				MeshData.Tangents.SetNormal(BottomLeftIndex, NormalCurrent);
				MeshData.Tangents.SetNormal(BottomRightIndex, NormalCurrent);
			}
		}
	}
	StaticProvider->UpdateSection(0, 0, MeshData);
}

// Called when the game starts or when spawned
void AClothActor::BeginPlay()
{
	Super::BeginPlay();
	
	ClothSimulation->NodeMass = NodeMass;
	ClothSimulation->StructuralElasticity = StructuralElasticity;
	ClothSimulation->ShearElasticity = ShearElasticity;
	ClothSimulation->FlexionElasticity = FlexionElasticity;
	ClothSimulation->Damping = Damping;
	ClothSimulation->TimeDilation = TimeDilation;
	ClothSimulation->Scale = Scale;
	ClothSimulation->GridSize = GridSize;
	ClothSimulation->CollisionActor = CollisionActor;

	ClothSimulation->GenerateCloth();
	GenerateMesh();
}

// Called every frame
void AClothActor::Tick(float DeltaTime)
{
	Super::Tick(DeltaTime);
	
	ClothSimulation->TimeDilation = TimeDilation;

	UpdateMesh();
}

