#include "ClothNode.h"
#include "ClothSpring.h"
#include "EngineUtils.h"

void FClothNode::AddSpring(const FClothSpringRef InSpring)
{
	Springs.Add(InSpring);
}

void FClothNode::Tick(float DeltaTime, UWorld* World, const FTransform& ParentTransform,
	TSubclassOf<AActor> CollisionType)
{
	const FVector3f LastPosition = Position;

	if (bFixed)
	{
		Speed = FVector3f::ZeroVector;
	}
	else
	{
		// Simple differential time step
		const FVector3f Acceleration = ActingForce / Mass;
		Speed += DeltaTime * Acceleration;
		Position += DeltaTime * Speed;
	}
	
	// Checks for collision only with our player ball actor.
	// Collision with other objects could be more easily achieved by i.e. each
	// cloth node being its own actor (or more accurately, a collider or a rigid body),
	// so that they can participate in Unreal's physics calculations
	// more easily (as a kinematic object).
	FRotator BackRotator = ParentTransform.Rotator().GetInverse();
	for (TActorIterator it(World, CollisionType); it; ++it) {
		FVector Location = it->GetActorLocation();
		Location = BackRotator.RotateVector(Location - ParentTransform.GetLocation());
		FVector3f Loc = static_cast<FVector3f>(Location);
		if (TestSphereIntersection(Loc, CollisionRadius)) {
			FVector3f ToNode = LastPosition - Loc;
			ChangeSpeedAcordingToIntersectedSphere(Loc, LastPosition);
			ToNode.Normalize();
			ToNode = ToNode * CollisionRadius * 1.05f;
			Position = Loc + ToNode;
			Position += Speed * DeltaTime;
		}
	}
}

void FClothNode::SetFixed(const bool bInFixed)
{
	this->bFixed = bInFixed;
}

FVector3f FClothNode::GetPosition() const
{
	return Position;
}

FVector3f FClothNode::GetSpeed() const
{
	return Speed;
}

void FClothNode::RecalculateForce()
{
	ActingForce = FVector3f::ZeroVector;

	for (FClothSpringRef s : Springs)
	{
		AddForce(s->GetForce(this));
	}
}

void FClothNode::AddForce(const FVector3f& InForce)
{
	this->ActingForce += InForce;
}

void FClothNode::ChangeSpeedAcordingToIntersectedSphere(const FVector3f& IntersectionPoint, const FVector3f& NodePosition)
{
	FVector3f ToNode = NodePosition - IntersectionPoint;
	ToNode.Normalize();
	Speed = FVector3f::VectorPlaneProject(Speed, ToNode);
}

bool FClothNode::TestSphereIntersection(const FVector3f& Location, const float Radius) const
{
	return FVector3f::DistSquared(Location, Position) < Radius * Radius * 1.05f;
}
