


#include "ClothSimulationComponent.h"

#include "ClothSpring.h"

// Sets default values for this component's properties
UClothSimulationComponent::UClothSimulationComponent()
{
	// Set this component to be initialized when the game starts, and to be ticked every frame.  You can turn these features
	// off to improve performance if you don't need them.
	PrimaryComponentTick.bCanEverTick = true;

	// ...
}


// Called when the game starts
void UClothSimulationComponent::BeginPlay()
{
	Super::BeginPlay();

	// ...
	
}


// Called every frame
void UClothSimulationComponent::TickComponent(float DeltaTime, ELevelTick TickType, FActorComponentTickFunction* ThisTickFunction)
{
	Super::TickComponent(DeltaTime, TickType, ThisTickFunction);
	
	FTransform Transform = GetOwner()->GetActorTransform();

	// Clamp time dilation to realtime at most.
	float Dilation = std::max(1.0f, TimeDilation);
	float Delta = DeltaTime / Dilation;

	UWorld* World = GetWorld();
	FVector3f Gravity = FVector3f::ZAxisVector * World->GetGravityZ() * Delta;

	// First recalculate new node positions based on their current calculated forces.
	for (FClothNodeRef Node : ClothNodes)
	{
		Node->Tick(Delta, World, Transform, CollisionActor);
	}
	
	// Then recalculate spring/damping forces for each cloth node.
	for (FClothNodeRef Node : ClothNodes)
	{
		Node->RecalculateForce();
		Node->AddForce(Gravity);
	}
}

bool UClothSimulationComponent::HasValidParameters()
{
	return GridSize > 0;
}

void UClothSimulationComponent::GenerateCloth()
{
	if (HasValidParameters())
	{
		ClothNodes.Empty();
		ClothNodes.Reserve((GridSize + 1) * (GridSize + 1));
		const FVector2f SectionSize = FVector2f(Scale / GridSize, Scale / GridSize);
		
		// Create cloth nodes (mass points)
		for (int32 Y = 0; Y <= GridSize; ++Y)
		{
			for (int32 X = 0; X <= GridSize; ++X)
			{
				FVector3f pos(0.f, Y * SectionSize.Y, X * -SectionSize.X);
				ClothNodes.Emplace(MakeShared<FClothNode>(pos, NodeMass));
			}
		}

		// Fix two corner points
		ClothNodes[0]->SetFixed(true);
		ClothNodes[ClothNodes.Num() - GridSize - 1]->SetFixed(true);
		
		// Add springs to nodes
		unsigned int index = 0;
		int32 NodeCountX = GridSize + 1;
		int32 NodeCountY = GridSize + 1;

		ClothSprings.Empty();
		
		for (int32 Y = 0; Y < NodeCountY; ++Y)
		{
			for (int32 X = 0; X < NodeCountX; ++X)
			{
				// Only add springs connecting this node to previous nodes.
				FClothNodeRef ThisNode = ClothNodes[Y * NodeCountX + X];

				// Left structural spring
				if (X > 0)
				{
					FClothNodeRef LeftNode = ClothNodes[Y * NodeCountX + X - 1];
					FClothSpringRef Spring = MakeShared<FClothSpring>(ThisNode, LeftNode, StructuralElasticity,
						FVector3f::Distance(ThisNode->GetPosition(), LeftNode->GetPosition()), Damping);
					ThisNode->AddSpring(Spring);
					LeftNode->AddSpring(Spring);
					ClothSprings.Add(Spring);
				}

				// Up structural spring
				if (Y > 0)
				{
					FClothNodeRef UpNode = ClothNodes[(Y - 1) * NodeCountX + X];
					FClothSpringRef Spring = MakeShared<FClothSpring>(ThisNode, UpNode, StructuralElasticity,
						FVector3f::Distance(ThisNode->GetPosition(), UpNode->GetPosition()), Damping);
					ThisNode->AddSpring(Spring);
					UpNode->AddSpring(Spring);
					ClothSprings.Add(Spring);
				}

				// Diagonal left up shear spring
				if (X > 0 && Y > 0)
				{
					FClothNodeRef UpLeftNode = ClothNodes[(Y - 1) * NodeCountX + X - 1];
					FClothSpringRef Spring = MakeShared<FClothSpring>(ThisNode, UpLeftNode, ShearElasticity,
						FVector3f::Distance(ThisNode->GetPosition(), UpLeftNode->GetPosition()), Damping);
					ThisNode->AddSpring(Spring);
					UpLeftNode->AddSpring(Spring);
					ClothSprings.Add(Spring);
				}

				// Diagonal right up shear spring
				if (Y > 0 && X < GridSize)
				{
					FClothNodeRef UpLeftNode = ClothNodes[(Y - 1) * NodeCountX + X + 1];
					FClothSpringRef Spring = MakeShared<FClothSpring>(ThisNode, UpLeftNode, ShearElasticity,
						FVector3f::Distance(ThisNode->GetPosition(), UpLeftNode->GetPosition()), Damping);
					ThisNode->AddSpring(Spring);
					UpLeftNode->AddSpring(Spring);
					ClothSprings.Add(Spring);
				}

				// Left flexion spring
				if (X > 1)
				{
					FClothNodeRef LeftNode = ClothNodes[Y * NodeCountX + X - 2];
					FClothSpringRef Spring = MakeShared<FClothSpring>(ThisNode, LeftNode, FlexionElasticity,
						FVector3f::Distance(ThisNode->GetPosition(), LeftNode->GetPosition()), Damping);
					ThisNode->AddSpring(Spring);
					LeftNode->AddSpring(Spring);
					ClothSprings.Add(Spring);
				}

				// Up flexion spring
				if (Y > 1)
				{
					FClothNodeRef UpNode = ClothNodes[(Y - 2) * NodeCountX+ X];
					FClothSpringRef Spring = MakeShared<FClothSpring>(ThisNode, UpNode, FlexionElasticity,
						FVector3f::Distance(ThisNode->GetPosition(), UpNode->GetPosition()), Damping);
					ThisNode->AddSpring(Spring);
					UpNode->AddSpring(Spring);
					ClothSprings.Add(Spring);
				}
			}
		}
	}
	else
	{
		UE_LOG(LogClass, Error, TEXT("Invalid cloth simulation parameters!"));
	}
}

FVector3f UClothSimulationComponent::GetNodePosition(int32 NodeIndex) const
{
	return ClothNodes[NodeIndex]->GetPosition();
}

